/*  Copyright 2008 Theo Berkau 
    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <wiiuse/wpad.h>
#include <ogcsys.h>
#include <gccore.h>
#include <fat.h>
#include <di/di.h>
#include "../cs0.h"
#include "../m68kcore.h"
#include "../peripheral.h"
#include "../vidsoft.h"
#include "../vdp2.h"
#include "../yui.h"
#include "perwii.h"
#include "sndwii.h"
#include "menu.h"

#include "../yui.h"

/* Wii includes */
#include "iowii.h"
//#include "syswii.h"

int YuiExec(void);

void Check_Bios_Exist(void);
void Check_SD_Init(void);
void Check_Read_Dir(void);
void scroll_filelist(s8 delta);
void print_filelist(void);
void scroll_cartlist(s8 delta);
void print_cartlist(void);

/* Constants */

static u32 *xfb[2] = { NULL, NULL };
int fbsel = 0;
static GXRModeObj *rmode = NULL;
volatile int done=0;
volatile int resetemu=0;
int running=1;
static void *_console_buffer = NULL;
void __console_init_ex(void *conbuffer,int tgt_xstart,int tgt_ystart,int tgt_stride,int con_xres,int con_yres,int con_stride);

SH2Interface_struct *SH2CoreList[] = {
&SH2Interpreter,
&SH2DebugInterpreter,
NULL
};

PerInterface_struct *PERCoreList[] = {
&PERDummy,
&PERWiiKeyboard,
&PERWiiClassic,
NULL
};

CDInterface *CDCoreList[] = {
&DummyCD,
&ISOCD,
NULL
};

SoundInterface_struct *SNDCoreList[] = {
&SNDDummy,
&SNDWII,
NULL
};

VideoInterface_struct *VIDCoreList[] = {
&VIDDummy,
&VIDSoft,
NULL
};

M68K_struct *M68KCoreList[] = {
&M68KDummy,
&M68KC68K,
NULL
};

static char bupfilename[512]="/SSROMS/bkram.bin";
static char biosfilename[512]="/SSROMS/bios.bin";
static char isofilename[512]="/SSROMS/game.cue";
static char cartfilename[11][512]={
 "",
 "/SSROMS/actionreplay.bin",
 "/SSROMS/bkram4M.bin",
 "/SSROMS/bkram8M.bin",
 "/SSROMS/bkram16M.bin",
 "/SSROMS/bkram32M.bin",
 "",
 "",
 "",
 "/SSROMS/ROM16M.bin",
 ""
};

extern int vdp2width, vdp2height;
int wii_width, wii_height;

#define FILES_PER_PAGE	5
#define ROMS_DIR	"/SSROMS"

struct file *filelist = NULL;
u32 nb_files = 0;
#define CART_TYPE_NUM 11

s32 selected = 0, start = 0;
s32 selectedcart = 5, startcart = 2;
int submenuflag = 0;

extern void sleep(u32);
void gotoxy(int x, int y);
void OnScreenDebugMessage(char *string, ...);
void DoMenu();
int GameExec();
int LoadCue();
int Settings();
int CartSet();
int CartSetExec();
int About();

int menuselect=1;
int setmenuselect=0;
int aboutmenuselect=3;
int numsetmenu;
int numcartmenu;
int numaboutmenu;

static BOOL flag_mount = FALSE;

void fat_remount()
{
   if(flag_mount)
   {
      fatUnmount("fat:/");
      flag_mount = FALSE;
   }
   fatInitDefault();
   flag_mount = TRUE;
}

void reset()
{
   resetemu=1;
}

void powerdown()
{
   done = 1;
}

int DVDStopMotor()
{
   static char dvdstr[] ATTRIBUTE_ALIGN(32) = "/dev/di";
   s32 fd;
   u8 buf[0x20];
   u8 outbuf[0x20];

   if ((fd = IOS_Open(dvdstr,0)) < 0)
      return 0;
	
   ((u32 *)buf)[0x00] = 0xE3000000;
   ((u32 *)buf)[0x01] = 0;
   ((u32 *)buf)[0x02] = 0;

   IOS_Ioctl(fd, buf[0], buf, 0x20, outbuf, 0x20);

   IOS_Close(fd);

   return 1;
}

int main(int argc, char **argv)
{
   WPAD_Init();
   PAD_Init();
   SYS_SetResetCallback(reset);
   SYS_SetPowerCallback(powerdown);
   //DVDStopMotor();
	
   fatInitDefault();

   DoMenu();

   exit(0);
   return 0;
}

int YuiExec()
{
   yabauseinit_struct yinit;
   int ret;
   FILE *fp;


   VIDEO_Init();

#ifndef HIGHRESO
   switch(VIDEO_GetCurrentTvMode()) 
   {
      case VI_NTSC:
         rmode = &TVNtsc240Ds;
	 break;
      case VI_PAL:
         rmode = &TVPal264Ds;
 	 break;
      case VI_MPAL:
	 rmode = &TVMpal480IntDf;
	 break;
      default:
         rmode = &TVNtsc240Ds;
	 break;
   }
#endif



#ifndef HIGHRESO
   // Allocate two buffers(may not be necessary)
   xfb[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
   xfb[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	
   VIDEO_Configure(rmode);
   VIDEO_ClearFrameBuffer (rmode, xfb[0], COLOR_BLACK);
   VIDEO_ClearFrameBuffer (rmode, xfb[1], COLOR_BLACK);
   VIDEO_SetNextFramebuffer(xfb[0]);
   VIDEO_SetBlack(FALSE);
   VIDEO_Flush();
   VIDEO_WaitVSync();
   if(rmode->viTVMode&VI_NON_INTERLACE) 
      VIDEO_WaitVSync();

#endif

   //free(_console_buffer);
   //if(_console_buffer != NULL) free(_console_buffer);
#if 0
   _console_buffer = malloc(10*10*VI_DISPLAY_PIX_SZ);
   __console_init_ex(_console_buffer,10,10,rmode->fbWidth*VI_DISPLAY_PIX_SZ,10,10,10*VI_DISPLAY_PIX_SZ);
#endif

   wii_width  = rmode->fbWidth / 2;
   wii_height = rmode->xfbHeight;

   memset(&yinit, 0, sizeof(yabauseinit_struct));
   yinit.percoretype = PERCORE_WIICLASSIC;
   yinit.sh2coretype = SH2CORE_INTERPRETER;
   yinit.vidcoretype = VIDCORE_SOFT;
   yinit.sndcoretype = SNDCORE_WII;
   yinit.cdcoretype = CDCORE_ISO;
   yinit.m68kcoretype = M68KCORE_C68K;
   yinit.carttype = selectedcart;
   yinit.regionid = REGION_AUTODETECT;
   if ((fp=fopen(biosfilename, "rb")) == NULL) {
      yinit.biospath = NULL;
   } else {
      yinit.biospath = biosfilename;
      fclose(fp);
   }
   yinit.cdpath = isofilename;
   yinit.buppath = bupfilename;
   yinit.mpegpath = NULL;
   yinit.cartpath = cartfilename[selectedcart];
   yinit.netlinksetting = NULL;
   yinit.flags = VIDEOFORMATTYPE_NTSC;

   // Hijack the fps display
   VIDSoft.OnScreenDebugMessage = OnScreenDebugMessage;


   if ((ret = YabauseInit(&yinit)) == 0)
   {
      EnableAutoFrameSkip();
      VIDEO_ClearFrameBuffer(rmode, xfb[fbsel], COLOR_BLACK);

      while(!done)
      {
         if (PERCore->HandleEvents() != 0)
            return -1;
         if (resetemu)
         {
            YabauseReset();
            resetemu = 0;
            SYS_SetResetCallback(reset);
         }
      }
      YabauseDeInit();
      done=0;
      resetemu=0;
   }
   else
   {
      while(!done)
         VIDEO_WaitVSync();
   }

   return 0;
}

void YuiErrorMsg(const char *string)
{
   if (strncmp(string, "Master SH2 invalid opcode", 25) == 0)
   {
      if (!running)
         return;
      running = 0;
      printf("%s\n", string);
   }
}

void YuiSwapBuffers()
{
   int i, j;
   u32 *curfb;
   u32 *buf;
   int index;

   fbsel ^= 1;
   curfb = xfb[fbsel];
   buf = dispbuffer;

   for (j = 0; j < wii_height; j++)
   {
      for (i = 0; i < wii_width; i++)
      {
         u8 y, cb, cr;
         u8 r, g, b;

         index = ((int)((double)j*(double)vdp2height/(double)wii_height)*vdp2width) + (int)((double)i*(double)vdp2width/(double)wii_width);

         r = buf[index] >> 24;
         g = buf[index] >> 16;
         b = buf[index] >> 8;

         y  = ( 257 * r + 504 * g +  98 * b +  16000) / 1000;
         cb = (-148 * r - 291 * g + 439 * b + 128000) / 1000; //b-y
         cr = ( 439 * r - 368 * g -  71 * b + 128000) / 1000; //r-y

         curfb[0] = (y << 24) | (cb << 16) | (y << 8) | cr;
         curfb++;
      }
   }

   VIDEO_SetNextFramebuffer (xfb[fbsel]);
   VIDEO_Flush ();
}

void gotoxy(int x, int y)
{
   printf("\033[%d;%dH", y, x);
}

void OnScreenDebugMessage(char *string, ...)
{
#if 0
   va_list arglist;

   va_start(arglist, string);
   gotoxy(0, 1);
   vprintf(string, arglist);
   printf("\n");
   gotoxy(0, 1);
   va_end(arglist);
#endif
}

int ClearMenu(const unsigned long *bmp)
{
   fbsel ^= 1;
   memcpy (xfb[fbsel], bmp, MENU_SIZE * 2);
   VIDEO_SetNextFramebuffer (xfb[fbsel]);
   VIDEO_Flush ();
   return 0;   
}

typedef struct
{
   char *name;
   int (*func)();
} menuitem_struct;

int MenuNone()
{
   return 0;
}

menuitem_struct menuitem[] = {
{ "Start emulation", YuiExec },
{ "Load ISO/CUE", LoadCue },
{ "Settings", Settings},
{ "About", About},
{ "Exit", NULL },
{ NULL, NULL }
};

menuitem_struct setmenuitem[] = {
{ "Cartridge", CartSet},
{ NULL, NULL }
};

menuitem_struct cartmenuitem[] = {
{ "None", CartSetExec},
{ "Action Relay", CartSetExec},
{ "Backup RAM  4MBIT", CartSetExec},
{ "Backup RAM  8MBIT", CartSetExec},
{ "Backup RAM 16MBIT", CartSetExec},
{ "Backup RAM 32MBIT", CartSetExec},
{ "DRAM  8MBIT", CartSetExec},
{ "DRAM 32MBIT", CartSetExec},
{ "NETLINK(fake)", MenuNone},
{ "ROM 16MBIT", CartSetExec},
{ "Japanese Modem", CartSetExec},
{ NULL, NULL }
};

menuitem_struct aboutmenuitem[] = {
{ "Yabause, a Sega Saturn Emulator", MenuNone},
{ VERSION, MenuNone},
{ "Original Yabause by Yabause Team", MenuNone},
{ "Return", NULL},
{ NULL, NULL }
};

void InitMenu()
{
   VIDEO_Init();
   switch(VIDEO_GetCurrentTvMode()) 
   {
      case VI_NTSC:
         rmode = &TVNtsc480IntDf;
	 break;
      case VI_PAL:
         rmode = &TVPal528IntDf;
 	 break;
      case VI_MPAL:
	 rmode = &TVMpal480IntDf;
	 break;
      default:
         rmode = &TVNtsc480IntDf;
	 break;
   }

   xfb[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
   xfb[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
   VIDEO_Configure(rmode);
   VIDEO_ClearFrameBuffer (rmode, xfb[0], COLOR_BLACK);
   VIDEO_ClearFrameBuffer (rmode, xfb[1], COLOR_BLACK);
   VIDEO_SetNextFramebuffer(xfb[0]);
   VIDEO_SetBlack(FALSE);
   VIDEO_Flush();
   VIDEO_WaitVSync();
   if(rmode->viTVMode&VI_NON_INTERLACE) 
      VIDEO_WaitVSync();
   //free(_console_buffer);
   if(_console_buffer != NULL) free(_console_buffer);
   _console_buffer = malloc(280*100*VI_DISPLAY_PIX_SZ);
   __console_init_ex(_console_buffer,180,200,rmode->fbWidth*VI_DISPLAY_PIX_SZ,280,100,280*VI_DISPLAY_PIX_SZ);
}

void DoMenu()
{
   //int menuselect=0;
   int i;
   int nummenu=0;
   expansion_t exp;
   s8 claX, claY;
   s8 gcX, gcY;

   InitMenu();

   for (i = 0; menuitem[i].name != NULL; i++)
      nummenu++;

   for (;;)
   {
      VIDEO_WaitVSync();
      WPAD_ScanPads();
      PAD_ScanPads();
      WPAD_Expansion(0, &exp);
      claX = classic_analog_val(&exp, true, false);
      claY = classic_analog_val(&exp, false, false);
      gcX = PAD_StickX(0);
      gcY = PAD_StickY(0);

      // Get Wii Remote/Keyboard/etc. presses
      if (WPAD_ButtonsDown(0) & WPAD_CLASSIC_BUTTON_UP ||
          WPAD_ButtonsDown(0) & WPAD_BUTTON_RIGHT ||
          PAD_ButtonsDown(0) & PAD_BUTTON_UP ||
          stick_up(claY, gcY))
      {
         switch (submenuflag) {
           case 0: //main menu
             menuselect--;
             if (menuselect < 0)
               menuselect = nummenu-1;
             break;
           case 1: //file selection menu
             scroll_filelist(-1);
             break;
           case 2: //setting menu
             setmenuselect--;
             if (setmenuselect < 0)
               setmenuselect = numsetmenu-1;
             break;
           case 21: //setting cart menu
             scroll_cartlist(-1);
             break;
         }
      }
      else if (WPAD_ButtonsDown(0) & WPAD_CLASSIC_BUTTON_DOWN ||
               WPAD_ButtonsDown(0) & WPAD_BUTTON_LEFT ||
               PAD_ButtonsDown(0) & PAD_BUTTON_DOWN ||
               stick_down(claY, gcY))
      {          
         switch (submenuflag) {
           case 0: //main menu
             menuselect++;
             if (menuselect == nummenu)
               menuselect = 0;
             break;
           case 1: //file selection menu
             scroll_filelist(1);
             break;
           case 2: //setting menu
             setmenuselect++;
             if (setmenuselect == numsetmenu)
               setmenuselect = 0;
             break;
           case 21: //setting cart menu
             scroll_cartlist(1);
             break;
         }
      }

      if (WPAD_ButtonsDown(0) & WPAD_CLASSIC_BUTTON_A ||
          WPAD_ButtonsDown(0) & WPAD_BUTTON_A ||
          WPAD_ButtonsDown(0) & WPAD_BUTTON_2 ||
          PAD_ButtonsDown(0) & PAD_BUTTON_A)
      {
          switch (submenuflag) {
             case 0: //main menu
               if (menuitem[menuselect].func)
               {
                 if (menuitem[menuselect].func())
                   return;
                 InitMenu();
               }
               else
                 return;
                 break;
             case 1: //file selection menu
                GameExec();
                break;
             case 2: //setting menu
               if (setmenuitem[setmenuselect].func)
               {
                 if (setmenuitem[setmenuselect].func())
                   return;
                 InitMenu();
               }
               else
                 return;
                 break;
             case 21: //setting cart menu
               if (cartmenuitem[selectedcart].func)
               {
                 if (cartmenuitem[selectedcart].func())
                   return;
                 InitMenu();
               }
               else
                 return;
                 break;
             case 3: //about menu
               if (aboutmenuitem[aboutmenuselect].func)
               {
                 if (aboutmenuitem[aboutmenuselect].func())
                   return;
                 InitMenu();
               }
               else
               {
                 submenuflag = 0;
                 menuselect = 1;
                 InitMenu();
                 //return;
               }
               break;
          }
      }

      // Draw menu
      ClearMenu(menu_bmp);

      printf("\033[2J");
      //printf("\033[2;0H");
      gotoxy(0, 0);
      switch (submenuflag) {
        case 0: //main menu
          // Draw menu items
          for (i = 0; i < nummenu; i++)
          {
            if (menuselect == i) {
              printf("\033[%d;%dm\033[35m", 30+9, 0);
              printf("%s\033[0m\n", menuitem[i].name);
            } else {
              printf("\033[%d;%dm", 30+7, 0);
              printf("%s\n", menuitem[i].name);
            }
          }
          break;
        case 1: //file selection menu
          print_filelist();
          break;
        case 2: //setting menu
          for (i = 0; i < numsetmenu; i++)
          {
            if (setmenuselect == i) {
              printf("\033[%d;%dm\033[35m", 30+9, 0);
              printf("%s\033[0m\n", setmenuitem[i].name);
            } else {
              printf("\033[%d;%dm", 30+7, 0);
              printf("%s\n", setmenuitem[i].name);
            }
          }
          break;
        case 21: //setting cart menu
          print_cartlist();
          break;
        case 3: //about menu
          for (i = 0; i < numaboutmenu; i++)
          {
            if (aboutmenuselect == i) {
              printf("\033[%d;%dm\033[35m", 30+9, 0);
              printf("%s\033[0m\n", aboutmenuitem[i].name);
            } else {
              printf("\033[%d;%dm", 30+7, 0);
              printf("%s\n", aboutmenuitem[i].name);
            }
          }
          break;
      }
   }
}

int GameExec()
{
   //submenuflag = 0;
   submenuflag = 21;
   printf("\033[2;0H");
   strcpy(isofilename, ROMS_DIR);
   strcat(isofilename, "/");
   strcat(isofilename, filelist[selected].filename);
   menuselect=0;
   return 0;
}
////////



void Check_SD_Init(void)
{
	s32 ret;
	
	//printf("\t[*] Initializing libfat...");
	fflush(stdout);
	ret = sd_init();
	if (!ret) {
		printf(" ERROR! Return to Wii Menu in 5s...");
		sleep(5);
		//Exit_ToWiiMenu();
		exit(0);
	}
	//printf(" OK!\n");
}

void Check_Read_Dir(void)
{
	s32 ret;
	
	//printf("\t[*] Reading SD card...");
	fflush(stdout);
	ret = sd_readdir(ROMS_DIR, &filelist, &nb_files);
	if (ret < 0) {
		printf(" ERROR! (ret = %d) Return to Wii Menu in 5s...", ret);
		sleep(5);
		//Exit_ToWiiMenu();
		exit(0);
	}
	//printf(" OK!\n");
}

void scroll_filelist(s8 delta)
{
	s32 index;

	/* No files */
	if (!nb_files)
		return;

	/* Select next entry */
	selected += delta;

	/* Out of the list? */
	if (selected <= -1)
		selected = nb_files - 1;
	if (selected >= nb_files)
		selected = 0;

	/* List scrolling */
	index = (selected - start);
	if (index >= FILES_PER_PAGE)
		start += (index - FILES_PER_PAGE) + 1;
	if (index <= -1)
		start += index;
}

void print_filelist(void)
{
	u32 cnt;

	//printf("\t[*] Available files on the SD card:\n\n");

	/* No files */
	if (!nb_files) {
		printf("No files available!\n");
		return;
	}

	for (cnt = start; cnt < nb_files; cnt++) {
		/* Files per page limit */
		if ((cnt - start) >= FILES_PER_PAGE)
			break;

		/* Selected file */
		if (cnt == selected) {
			//printf("\t\t>> ");
                      printf("\033[%d;%dm\033[35m", 30+9, 0);
		      printf("%s\033[0m\n", filelist[cnt].filename);
		} else {
			//printf("\t\t   ");
		      printf("\033[%d;%dm", 30+7, 0);
		      printf("%s\n", filelist[cnt].filename);
		}
		fflush(stdout);

	}
}

void scroll_cartlist(s8 delta)
{
        s32 index;

        /* Select next entry */
        selectedcart += delta;

        /* Out of the list? */
        if (selectedcart <= -1)
                selectedcart = CART_TYPE_NUM - 1;
        if (selectedcart >= CART_TYPE_NUM)
                selectedcart = 0;

        /* List scrolling */
        index = (selectedcart - startcart);
        if (index >= FILES_PER_PAGE)
                startcart += (index - FILES_PER_PAGE) + 1;
        if (index <= -1)
                startcart += index;
}

void print_cartlist(void)
{
        u32 cnt;

        for (cnt = startcart; cnt < CART_TYPE_NUM; cnt++) {
                /* Files per page limit */
                if ((cnt - startcart) >= FILES_PER_PAGE)
                        break;

                /* Selected file */
                if (cnt == selectedcart) {
                      printf("\033[%d;%dm\033[35m", 30+9, 0);
                      printf("%s\033[0m\n", cartmenuitem[cnt].name);
                } else {
                      printf("\033[%d;%dm", 30+7, 0);
                      printf("%s\n", cartmenuitem[cnt].name);
                }
                fflush(stdout);
        }
}

////////
int LoadCue()
{
   submenuflag = 1;
   printf("\033[2;0H");
   gotoxy(0, 0);
   Check_Read_Dir();

   return 0;
}


int Settings()
{
   int i;
   submenuflag = 2;
   printf("\033[2;0H");
   gotoxy(0, 0);
   numsetmenu=0;
   for (i = 0; setmenuitem[i].name != NULL; i++)
      numsetmenu++;

   return 0;
}

int CartSet()
{
   int i;
   submenuflag = 21;
   printf("\033[2;0H");
   gotoxy(0, 0);
   numcartmenu=0;
   for (i = 0; cartmenuitem[i].name != NULL; i++)
      numcartmenu++;

   return 0;
}

int CartSetExec()
{
   submenuflag = 0;
   menuselect=0;
   return 0;
}

int About()
{
   int i;
   submenuflag = 3;
   printf("\033[2;0H");
   gotoxy(0, 0);
   numaboutmenu=0;
   for (i = 0; aboutmenuitem[i].name != NULL; i++)
      numaboutmenu++;

   return 0;
}

